#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdlib.h>
#include <time.h>
#include <gtk/gtk.h>
#include "socket.h"
#include "log.h"
#include "rcfile.h"
#include "callbacks.h"
#include "main.h"
#include "interface.h"
#include "support.h"

static void get_config(char **server, char **port, char **username, 
		char **password);
static void get_config_main(const char *lookup, char **s);
static void set_config(const char *server, const char *port, 
		const char *username, const char *password);
static void set_config_main(const char *lookup, const char *s);
static void add_log(GtkText *status, const char *msg);
static gint keepalive(gpointer user_data);
static void print_sock(gpointer user_data, gint source, 
		GdkInputCondition condition);
static void dial_add_monitors(void);
static void dial_remove_monitors(void);

int sok = -1;
gint tag_keepalive;
gint tag_monitor;
gboolean log_window_visible = FALSE;

void get_config(char **server, char **port, char **username, char **password)
{
	get_config_main("server", server);
	get_config_main("port", port);
	get_config_main("username", username);
	get_config_main("password", password);
	return;
}

void get_config_main(const char *lookup, char **s)
{
	GtkWidget *entry;
	entry = lookup_widget(GTK_WIDGET(config_window), lookup);
	*s = gtk_entry_get_text(GTK_ENTRY(entry));
	return;
}

void set_config(const char *server, const char *port, const char *username, 
		const char *password)
{
	set_config_main("server", server);
	set_config_main("port", port);
	set_config_main("username", username);
	set_config_main("password", password);
	return;
}

void set_config_main(const char *lookup, const char *s)
{
	GtkWidget *entry;
	if (!s)
		return;
	entry = lookup_widget(GTK_WIDGET(config_window), lookup);
	gtk_entry_set_text(GTK_ENTRY(entry), s);
	return;
}

void dial_log_raw(const char *msg)
{
	GtkWidget *status;
	status = lookup_widget(log_window, "status");
	add_log(GTK_TEXT(status), msg);
}

/* Taken from
 * http://www.gnome.org/mailing-lists/archives/gtk-app-devel-list/1999-May/0327.shtml
 * written by Noonian Soong */
void add_log(GtkText *status, const char *msg)
{
	GtkAdjustment *adj;
	gboolean snap = FALSE;

	adj = status->vadj;
	if (adj->upper - adj->page_size == adj->value)
		snap = TRUE;
	gtk_text_freeze(status);
	gtk_text_insert(status, NULL, NULL, NULL, msg, -1);
	gtk_text_thaw(status);
	if (snap)
		gtk_adjustment_set_value( adj, adj->upper - adj->lower -
				adj->page_size );
}

void dial_status(const char *msg)
{
	GtkWidget *statusbar;
	statusbar = lookup_widget(GTK_WIDGET(main_window), "statusbar");
	gtk_statusbar_push(GTK_STATUSBAR(statusbar), status_context_id, msg);
}

void init_status(void)
{
	GtkWidget *statusbar;
	statusbar = lookup_widget(GTK_WIDGET(main_window), "statusbar");
	status_context_id = gtk_statusbar_get_context_id(
			GTK_STATUSBAR(statusbar), "connection status");
	return;
}

void dial_setfrom_rcfile(char *server, char *port, char *username, 
		char *password)
{
	set_config(server, port, username, password);
	free(server); free(port); free(username); free(password);
}

void on_button_connect_clicked(GtkButton *button, gpointer user_data)
{
	char *server, *port, *username, *password;
	get_config(&server, &port, &username, &password);
	if (sok!=-1)
		disconnect();
	sok = dial_sock_connect(server, port, username, password);
	if (sok==-1) {
		dial_status("Error connecting");
		return;
	}
	dial_add_monitors();
	return;
}

void dial_add_monitors(void)
{
	/* every 30 seconds, run keepalive() */
	tag_keepalive = gtk_timeout_add(30000, keepalive, NULL);
	/* everytime there is data on sok, run print_sock() */
	tag_monitor = gdk_input_add(sok, GDK_INPUT_READ, print_sock, NULL);
	return;
}

void dial_remove_monitors(void)
{
	gtk_timeout_remove(tag_keepalive);
	gdk_input_remove(tag_monitor);
	return;
}

gint keepalive(gpointer user_data)
{
	dial_sock_keepalive(sok);
	return TRUE;
}

void print_sock(gpointer user_data, gint source, GdkInputCondition condition)
{
	dial_sock_logdata(sok);
	return;
}

void on_button_disconnect_clicked(GtkButton *button, gpointer user_data)
{
	disconnect();
}

void disconnect(void)
{
	if (sok==-1)
		return; /* already down */
	dial_log("ordering connection down\n");
	if (dial_sock_disconnect(sok))
		dial_status("Disconnect error");
	else
		dial_status("Disconnected");
	dial_remove_monitors();
	sok=-1;
	return;
}

void on_button_config_clicked(GtkButton *button, gpointer user_data)
{
	gtk_widget_show(config_window);
}

void on_button_viewlog_clicked(GtkButton *button, gpointer user_data)
{
	if (log_window_visible)
		gtk_widget_hide(log_window);
	else
		gtk_widget_show(log_window);
	log_window_visible = !log_window_visible;
	return;
}

void on_button_config_save_clicked(GtkButton *button, gpointer user_data)
{
	char *server, *port, *username, *password;
	GtkWidget *save_password = lookup_widget(GTK_WIDGET(config_window), 
			"save_password");

	get_config(&server, &port, &username, &password);
	if (!GTK_TOGGLE_BUTTON(save_password)->active)
		password = NULL;

	if (dial_rcfile_write(server, port, username, password))
		dial_status("Saving config failed");
	else
		dial_status("Config saved");
	gtk_widget_hide(config_window);
}

void on_button_config_nosave_clicked(GtkButton *button, gpointer user_data)
{
	dial_status("Config not saved");
	gtk_widget_hide(config_window);
}

void on_button_quit_clicked(GtkButton *button, gpointer user_data)
{
	gtk_exit(0);
}

gboolean on_main_deleted (GtkWidget *widget, GdkEvent *event, 
		gpointer user_data)
{
	gtk_exit(0);
	return FALSE;
}

gboolean on_config_window_delete (GtkWidget *widget, GdkEvent *event,
                                        gpointer user_data)
{
	dial_status("Config not saved");
	gtk_widget_hide(config_window);
	return TRUE;
}

gboolean on_log_window_delete (GtkWidget *widget, GdkEvent *event,
                                        gpointer user_data)
{
	gtk_widget_hide(log_window);
	log_window_visible = FALSE;
	return TRUE;
}
