#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/param.h>
#include "rcfile.h"
#include "callbacks.h"
#include "log.h"

static char *dial_rcfile;

#define MAXLINE 1024

static int get_settings(FILE *fp, char **server, char **port, char **username,
		char **password);
static int getline(FILE *fp, char **argument);
static int getline_keyword(const char *buf);
static int check_rcfile_perms(void);

enum {
	RC_INVALID,
	RC_EOF,
	RC_FATAL,
	RC_COMMENT,
	RC_PORT,
	RC_SERVER,
	RC_USERNAME,
	RC_PASSWORD
};

int dial_rcfile_setname(void)
{
	if (!(dial_rcfile = malloc(MAXPATHLEN))) {
		dial_log("malloc failed");
		return -1;
	}
	snprintf(dial_rcfile, MAXPATHLEN, "%s/.gcdialrc", getenv("HOME"));
	return 0;
}

int dial_rcfile_read(void)
{
	FILE *fp;
	char *server, *port, *username, *password;
	server = port = username = password = NULL;

	if (!(fp = fopen(dial_rcfile, "r"))) {
		dial_log_err("opening rcfile failed");
		return -1;
	}

	if (get_settings(fp, &server, &port, &username, &password)) {
		fclose(fp);
		return -1;
	}
	fclose(fp);
	dial_setfrom_rcfile(server, port, username, password);
	return 0;
}

int dial_rcfile_write(char *server, char *port, char *username, char *password)
{
	FILE *fp;

	umask(077);
	if (!(fp = fopen(dial_rcfile, "w"))) {
		dial_log("cannot open '%s' for writing", dial_rcfile);
		return -1;
	}
	if (server && *server!='\0')
		fprintf(fp, "server %s\n", server);
	if (port && *port!='\0')
		fprintf(fp, "port %s\n", port);
	if (username && *username!='\0')
		fprintf(fp, "username %s\n", username);
	if (password && *password!='\0') {
		if (check_rcfile_perms()) {
			dial_log("open permissions, not saving password\n");
			fclose(fp);
			return -1;
		}
		fprintf(fp, "password %s\n", password);
	}
	fclose(fp);
	return 0;
}

int get_settings(FILE *fp, char **server, char **port, char **username,
		char **password)
{
	int ret;
	char *arg;
	int line = 0;
	int error = 0;

	if (RC_FATAL==(ret = getline(fp, &arg)))
		return -1;
	while (ret!=RC_EOF) {
		++line;
		if (ret==RC_INVALID) {
			dial_log("invalid keyword in rcfile line %d\n", line);
			error = -1;
		} else if (!arg) {
			switch(ret) {
			case RC_COMMENT : break;
			default	: 
				  dial_log("keyword requires an argument in "
					  	  "rcfile line %d\n", line);
				  error = -1;
				  break;
			}
		} else {
			switch(ret) {
			case RC_SERVER: 
				free(*server);
				*server = arg;
				break;
			case RC_PORT: 
				free(*port);
				*port = arg;
				break;
			case RC_USERNAME:
				free(*username);
				*username = arg;
				break;
			case RC_PASSWORD:
				if (check_rcfile_perms()) {
					dial_log("warning: gcdial permissions "
							"are too open for "
							"password\n");
					error = -1;
				} else {
					free(*password);
					*password = arg;
				}
				break;
			default: 
				dial_log("keyword not allowed argument in "
						"rcfile line %d\n", line);
				error = -1;
				break;
			} /* switch(ret) */
		} /* arg !NULL */
		if (RC_FATAL==(ret = getline(fp, &arg)))
			return -1;
	} /* while(ret!=RC_EOF) */
	if (error)
		return -1; /* having printed all other errors */
	return 0;
}

int getline(FILE *fp, char **argument)
{
	char *buf;
	int com_len;
	int ret;
	char *pt;

	if (!(buf = malloc(MAXLINE)))
		return RC_FATAL;
	if (!fgets(buf, MAXLINE, fp))
		return RC_EOF;
	if ((com_len = strchr(buf, ' ') - buf) > 0) {
		/* argument */
		if (!(*argument = strdup(buf+com_len+1))) {
			dial_log_err("malloc failed");
			return RC_FATAL;
		}
		for (pt=*argument; *pt; ++pt);
		*(pt-1) = '\0'; /* bye bye \n */
		/* command */
		*(buf+com_len) = '\0';
	} else {
		/* command without an argument */
		*argument = NULL;
		for (pt=buf; *pt; ++pt);
		*(pt-1) = '\0'; /* bye bye \n */
	}
	if (*buf=='#') {
		free(buf);
		*argument = NULL;
		return RC_COMMENT;
	}
	ret = getline_keyword(buf);
	free(buf);
	return ret;
}

int getline_keyword(const char *buf)
{
	if (!strcasecmp(buf, "server"))
		return RC_SERVER;
	else if (!strcasecmp(buf, "port"))
		return RC_PORT;
	else if (!strcasecmp(buf, "username"))
		return RC_USERNAME;
	else if (!strcasecmp(buf, "password"))
		return RC_PASSWORD;
	else
		return RC_INVALID;
}

/* FIXME: should use file descriptor */
int check_rcfile_perms(void)
{
	struct stat buf;
	if (-1==stat(dial_rcfile, &buf))
		return -1;
	if ((buf.st_mode & ~S_IWUSR) != (S_IFREG | S_IRUSR))
		return -1;
	return 0;
}
