#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/time.h>
#include <signal.h>
#include <netdb.h>
#include "socket.h"
#include "log.h"
#include "callbacks.h"

#define CONNECT_TIMEOUT_SECS 10
#define MAXLINE 1024

static struct hostent *sock_gethostname(const char *server);
static int sock_connect_meat(int sok, struct sockaddr *addr);
static void sock_error(int err);
static int sock_send(int sok, struct timeval *tv, void *msg, size_t len_left);
static int sock_getstring(char *msg, int maxline, int sok, struct timeval *tv);
static int sock_sendline(int sok, char *msg);
static int sock_okay(int sok, struct timeval *tv); 

enum {
	SOK_SUCCESS,
	SOK_FAILED,
	SOK_TIMEOUT,
	SOK_TOOLONG
};

void dial_sock_keepalive(int sok)
{
	if (sock_sendline(sok, "ALIVE\n"))
		disconnect();
	return;
}

void dial_sock_logdata(int sok)
{
	struct timeval tv;
	tv.tv_usec = 0;
	tv.tv_sec = SOCKET_TIMEOUT_SECS;
	if (-1==sock_okay(sok, &tv))
		disconnect();
	return;
}
int dial_sock_disconnect(int sok)
{
	int err;
	err = sock_sendline(sok, "DISCONNECT\nQUIT\n");
	(void)shutdown(sok, 2);
	close(sok);
	return err;
}
	
int dial_sock_connect(const char *server, const char *port, 
		const char *username, const char *password)
{
	struct sockaddr_in addr;
	struct hostent *host;
	char *buf;
	int sok;

	if (!*server || *server=='\0') {
		dial_log("need to specify a server\n");
		return -1;
	} else if (!*port || *port=='\0') {
		dial_log("need to specify a port\n");
		return -1;
	} else if (!*username || *username=='\0') {
		dial_log("need to specify a username\n");
		return -1;
	} else if (!*password || *password=='\0') {
		dial_log("need to specify a password\n");
		return -1;
	}
	if (!(host = sock_gethostname(server)))
		return -1;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(atoi(port));
	addr.sin_addr.s_addr = inet_addr(inet_ntoa(*((struct in_addr *)
					host->h_addr)));
	memset(&addr.sin_zero, 0, sizeof(addr.sin_zero));
	if (-1==(sok = socket(AF_INET, SOCK_STREAM, 0))) {
		dial_log_err("creating socket failed");
		return -1;
	}
	if (-1==fcntl(sok, F_SETFL, O_NONBLOCK)) {
		dial_log_err("setting socket non-blocking failed");
		return -1;
	}
	if (sock_connect_meat(sok, (struct sockaddr *)&addr))
		return -1;
	dial_log("ordering connection up\n");
	if (!(buf = malloc(MAXLINE))) {
		dial_log_err("malloc failed");
		return -1;
	}
	snprintf(buf, MAXLINE, "USER %s\n", username);
	if (sock_sendline(sok, buf)) {
		close(sok);
		free(buf);
		return -1;
	}
	snprintf(buf, MAXLINE, "PASS %s\n", password);
	if (sock_sendline(sok, buf)) {
		close(sok);
		free(buf);
		return -1;
	}
	free(buf);
	if (sock_sendline(sok, "CONNECT\n")) {
		close(sok);
		return -1;
	}
	return sok;
}

int dial_sock_sigsetup(void)
{
	struct sigaction ignore;
	memset(&ignore, 0, sizeof(ignore));
	ignore.sa_handler = SIG_IGN;
	return sigaction(SIGPIPE, &ignore, NULL);
}

struct hostent *sock_gethostname(const char *server)
{
	struct hostent *host;
	if (!(host = gethostbyname(server))) {
		switch(h_errno) {
		case HOST_NOT_FOUND: 
			dial_log("server not found\n");
			break;
		case NO_ADDRESS:
			dial_log("server has no IP address\n");
			break;
		default:
			dial_log("looking up server IP failed\n");
			break;
		}
	}
	return host;
}

int sock_connect_meat(int sok, struct sockaddr *addr)
{
	fd_set sdfd;
	struct timeval tv;
	int ret;

	ret = connect(sok, addr, sizeof(struct sockaddr));
	if (!ret)
		return 0;
	if (errno!=EINPROGRESS) {
		dial_log_err("connecting to server failed");
		return -1;
	}
	FD_ZERO(&sdfd);
	FD_SET(sok, &sdfd);
	memset(&tv, 0, sizeof(struct timeval));
	tv.tv_sec = CONNECT_TIMEOUT_SECS;

	ret = select(sok+1, NULL, &sdfd, NULL, &tv);
	if (ret==-1) {
		dial_log_err("connecting to server failed");
		return -1;
	}
	if (ret==0) {
		dial_log("connecting to server failed: timeout\n");
		return -1;
	}
	if (sock_sendline(sok, "ALIVE\n")) {
		close(sok);
		return -1; /* write fails if connect() failed */
	}
	return 0;
}

int sock_sendline(int sok, char *msg)
{
	struct timeval tv;
	int ret;

	if (sok==-1)
		return -1;
	tv.tv_usec = 0;
	tv.tv_sec = SOCKET_TIMEOUT_SECS;

	ret =  sock_send(sok, &tv, msg, strlen(msg));
	if (ret!=SOK_SUCCESS) {
		sock_error(ret);
		return -1;
	}
	do {
		ret = sock_okay(sok, &tv);
	} while (ret==1);
	if (ret==0)
		return 0;
	return -1;
}
			  
int sock_okay(int sok, struct timeval *tv)
{
	char *buf;
	int ret;

	if (sok==-1)
		return -1;
	if (!(buf = malloc(MAXLINE+1))) {
		dial_log_err("malloc failed");
		return -1;
	}
	ret = sock_getstring(buf, MAXLINE, sok, tv);
	if (ret!=SOK_SUCCESS) {
		free(buf);
		sock_error(ret);
		return -1;
	}
	if (!strcmp(buf, "OKAY")) {
		free(buf);
		return 0; /* great */
	}
	dial_log("->%s\n", buf);
	if (*buf!='#') {
		free(buf);
		return -1;
	}
	if (!strcmp(buf, "#Connecting")) {
		dial_status("Connecting to ISP");
	} else if (!strcmp(buf, "#Connected")) {
		dial_status("Connected to ISP");
	} else if (!strcmp(buf, "#Connect failed")) {
		dial_status("Failed to connect to ISP");
	}
	free(buf);
	return 1; /* try again */
}

void sock_error(int err)
{
	switch(err) {
	case SOK_FAILED: 
		dial_log("connection reset by peer\n");
       		break;
	case SOK_TIMEOUT:
		dial_log("network timeout\n");
      		break;
	case SOK_TOOLONG:
		dial_log("garbage received from peer\n");
      		break;
	}
	return;
}

int sock_send(int sok, struct timeval *tv, void *msg, size_t len_left)
{
	fd_set sokset;
	struct timeval tvbak;
	int ret;

	FD_ZERO(&sokset);
	FD_SET(sok, &sokset);

	while (len_left > 0) {
		do {
			ret = write(sok, msg, len_left);
		} while (ret==-1 && errno==EINTR);
		if (ret==-1) {
			if (errno!=EAGAIN)
				return SOK_FAILED;
			if (tv!=NULL)
				memcpy(tv, &tvbak, sizeof(*tv));
			do {
				ret = select(sok+1, NULL, &sokset, NULL, tv);
			} while (ret==-1 && errno==EINTR);
			switch(ret) {
				case -1: return SOK_FAILED;
				case  0: return SOK_TIMEOUT;
				default: break;
			}
		} else
			len_left -= ret;
	}
	return SOK_SUCCESS;
}

int sock_getstring(char *msg, int maxline, int sok, struct timeval *tv)
{
	int numchars;
	int ret;
	struct timeval tvbak;
	fd_set reading;

	if (tv!=NULL)
		memcpy(&tvbak, tv, sizeof(tvbak));
	numchars = 0;
	while (numchars < maxline) {
		do {
			ret = read(sok, msg+numchars, 1);
		} while (ret==-1 && errno==EINTR);
		if (ret==-1) {
			if (errno!=EAGAIN) 
				return SOK_FAILED;
			FD_ZERO(&reading);
			FD_SET(sok, &reading);
			if (tv!=NULL)
				memcpy(tv, &tvbak, sizeof(*tv));
			do {
				ret = select(sok+1, &reading, NULL, NULL, tv);
			} while (ret==-1 && errno==EINTR);
			if (ret==-1)
				return SOK_FAILED;
			else if (ret==0)
				return SOK_TIMEOUT;
		} else if (ret==0) {
			return SOK_FAILED; /* EOF */
		} else { 
			if ('\n'==*(msg+numchars)) {
				/* telnet likes to end with '\r\n' */
				if ((numchars>0) && ('\r'==*(msg+numchars-1)))
					numchars--;
				*(msg+numchars) = '\0';
				return SOK_SUCCESS;
			}
			numchars++;
		}
	}
	/* string ending in \r possible */
	*(msg+numchars) = '\0';
	return SOK_TOOLONG;
}
